/******************************************************************************
*									      *
*	File:     dgpsip.c						      *
*	Author:   Wolfgang Rupprecht <wolfgang@wsrcc.com>	              *
*	Created:  Sun Jan 24 20:36:39 PST 1999				      *
*	Contents: DGPS over ip -- get RTCM signals from a remote tcp source   *
*									      *
*	$Id: dgpsip.c,v 1.29 1999/02/23 05:29:28 wolfgang Exp $
******************************************************************************/

/*
 *  Copyright (c) 1999 Wolfgang Rupprecht
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  			    NO WARRANTY
 *
 *    BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO
 *  WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE
 *  LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
 *  HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT
 *  WARRANTY OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING, BUT
 *  NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS TO THE
 *  QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
 *  PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
 *  SERVICING, REPAIR OR CORRECTION.
 *
 *    IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
 *  WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY
 *  MODIFY AND/OR REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE
 *  LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL,
 *  INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR
 *  INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
 *  DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU
 *  OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY
 *  OTHER PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN
 *  ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>		/* for serial hacking */
#include <fcntl.h>
#include <sys/termios.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <math.h>

#include "dgpsip.h"
#include "rtcm.h"

#define BUFSZ     (4 * 1024)
#define NMEABUFSZ (4 * 1024)

int             verbose = 0;

void            usage(void);
int             openserial(u_char * tty, int blocksz, int ttybaud);
void            fullwrite(int fd, u_char * buf, int cnt);
int             tcpopen(u_char * host, u_char * servp);

u_char         *nowarranty = "\n\
  BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY\n\
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN\n\
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES\n\
PROVIDE THE PROGRAM \" AS IS \" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESS\n\
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF\n\
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS\n\
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE\n\
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,\n\
REPAIR OR CORRECTION.\n\
\n\
  IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING\n\
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR\n\
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,\n\
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING\n\
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED\n\
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY\n\
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER\n\
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE\n\
POSSIBILITY OF SUCH DAMAGES.\n\n\
";

int
main(int argc,
     char **argv)
{
    u_char         *ttyout = "/dev/gps",
                   *ttynme = NULL,
                   *dgpshost = "dgps.wsrcc.com",
                   *dgpsport = "rtcm-sc104",
                    buf[BUFSZ],
                    nmeabuf[NMEABUFSZ];
    int             cnt,
                    c,
                    i,
                    ncnt,
                    rd,
                    dflag = 0,
                    infd = -1,
                    outfd = -1,
                    nmefd = -1,
                    nmebaud = 4800,
                    slp = 1,
                    wflag = 0,
                    nmeaind = 0,
                    cutoff = 16,
                    pts = 0,
                    ttybaud = 4800;
    double          x,
                    y,
                    h,
                    sdevlat,
                    sdevlon,
                    sdevheight;
    struct llpoint  point,
                    sum = {0.0, 0.0, 0.0},
                    sumdif = {0.0, 0.0, 0.0};
    FILE           *posf = NULL;

    while ((c = getopt(argc, argv, "c:dh:n:N:o:p:s:v:w")) != EOF) {
	switch (c) {
	case 'c':		/* capture GPS output to file */
	    if ((posf = fopen(optarg, "a")) == NULL) {
		perror("open logfile");
		exit(1);
	    }
	    break;

	case 'd':		/* decode */
	    dflag = 1;
	    ttyout = NULL;
	    break;

	case 'h':		/* dgps host */
	    dgpshost = optarg;
	    break;

	case 'p':		/* dgps port */
	    dgpsport = optarg;
	    break;

	case 'o':		/* gps serial ttyout */
	    ttyout = optarg;
	    break;

	case 'n':		/* gps serial NMEA input */
	    ttynme = optarg;
	    break;

	case 'N':		/* serial NMEA input speed */
	    nmebaud = atoi(optarg);
	    break;

	case 's':		/* serial ttyout speed */
	    ttybaud = atoi(optarg);
	    break;

	case 'v':		/* verbose */
	    verbose = atoi(optarg);
	    break;

	case 'w':		/* turn off nowarranty printf at startup */
	    wflag = 1;
	    break;

	case '?':
	default:
	    usage();
	    break;
	}
    }

    if (verbose >= 1)
	fprintf(stderr, "%s %s\n", PACKAGE, VERSION);

    if (wflag == 0) {
	fprintf(stderr, "%s", nowarranty);
    }
    if (ttyout) {
	outfd = openserial(ttyout, 1, ttybaud);
	if (outfd < 0) {
	    perror("open tty");
	    exit(1);
	}
	if (!ttynme || (strcmp(ttyout, ttynme) == 0)) {
	    nmefd = outfd;
	} else {
	    nmefd = openserial(ttynme, 1, nmebaud);
	    if (nmefd < 0) {
		perror("open NMEA tty");
		exit(1);
	    }
	}
    }
    while (infd < 0) {

	initialise();		/* init rtcm decoder */

	infd = tcpopen(dgpshost, dgpsport);
	if (infd < 0) {
	    fprintf(stderr,
	       "Can't open network input stream to %s:%s sleeping %d sec\n",
		    dgpshost, dgpsport, slp);
	    sleep(slp);
	    slp *= 2;
	    if (slp > 60)
		slp = 60;
	} else {
	    {
		u_char          hn[256];

		gethostname(hn, sizeof(hn));
		snprintf(buf, sizeof(buf), "HELO %s %s %s%s\r\n", hn, PACKAGE,
			 VERSION, (outfd < 0) ? "\r\nB":"");
		fullwrite(infd, buf, strlen(buf));
	    }
	    slp = 1;
	    while ((cnt = read(infd, buf, sizeof(buf))) > 0) {
		if (dflag) {
		    for (i = 0; i < cnt; i++) {
			new_byte(buf[i]);
		    }
		}
		if (outfd >= 0) {
		    fullwrite(outfd, buf, cnt);
#ifndef DONT_READ_NMEA
		    ncnt = read(nmefd, nmeabuf + nmeaind, NMEABUFSZ - nmeaind);
		    if (ncnt > 0) {
			if (posf != NULL) {
			    fwrite(nmeabuf + nmeaind, ncnt, (size_t) 1, posf);
			}
			nmeaind += ncnt;
			while ((rd = nmeaRead(nmeabuf, nmeaind, &point)) > 0) {
			    if (point.fix > 0) {
				sum.lat += point.lat;
				sum.lon += point.lon;
				sum.height += point.height;
				sum.hdop += point.hdop;
				sum.numsats += point.numsats;
				pts++;

				x = point.lat - (sum.lat / pts);
				sumdif.lat += x * x;
				y = point.lon - (sum.lon / pts);
				sumdif.lon += y * y;
				h = point.height - (sum.height / pts);
				sumdif.height += h * h;

				if (pts >= cutoff) {
				    u_char          rptbuf[BUFSZ];
				    cutoff *= 2;

				    sdevlat = sqrt(sumdif.lat / (pts - 1));
				    sdevlon = sqrt(sumdif.lon / (pts - 1));
				    sdevheight = sqrt(sumdif.height / (pts - 1));
				    snprintf(rptbuf, BUFSZ,
					     "R %0.8f %0.8f %0.2f -- %d %0.2e %0.2e %0.2e %0.1f %0.1f\n",
					     sum.lat / pts, sum.lon / pts,
					     sum.height / pts, pts,
					     sdevlat, sdevlon, sdevheight,
					     sum.hdop / pts,
					     sum.numsats / pts);

				    if (verbose >= 2) {
					printf("%s", rptbuf);
				    }
#if 1				/* only send 2 digit on the wire */
				    snprintf(rptbuf, BUFSZ,
					     "R %0.2f %0.2f %0.2f -- %d %0.2e %0.2e %0.2e %0.1f %0.1f\n",
					     sum.lat / pts, sum.lon / pts,
					     sum.height / pts, pts,
					     sdevlat, sdevlon, sdevheight,
					     sum.hdop / pts,
					     (double) (sum.numsats) / pts);
#endif
				    fullwrite(infd, rptbuf,
					      strlen(rptbuf));
				}
			    }
			    nmeaind -= rd;
			    memmove(nmeabuf, nmeabuf + rd, nmeaind);
			}
		    }
#endif
		}
		if (verbose >= 10) {
		    fprintf(stderr, "Read %d bytes\n", cnt);
		} else if (verbose >= 3) {
		    fprintf(stderr, ".");
		}
	    }
	    close(infd);
	    infd = -1;
	    fprintf(stderr,
		    "Network input stream to %s:%s Closed.  Reopening.\n",
		    dgpshost, dgpsport);
	}
    }
    exit(0);
}

int
openserial(u_char * tty, int blocksz, int ttybaud)
{
    int             fd;
    struct termios  termios;

    fd = open(tty, O_RDWR | O_NONBLOCK);
    if (fd < 0) {
	perror("open");
	return (-1);
    }
    if (tcgetattr(fd, &termios) < 0) {

	perror("tcgetattr");
	return (-1);
    }
    termios.c_iflag = 0;
    termios.c_oflag = 0;	/* (ONLRET) */
    termios.c_cflag = CS8 | CLOCAL | CREAD;
    termios.c_lflag = 0;
    {
	int             cnt;

	for (cnt = 0; cnt < NCCS; cnt++)
	    termios.c_cc[cnt] = -1;
    }
    termios.c_cc[VMIN] = blocksz;
    termios.c_cc[VTIME] = 2;

#if (B4800 != 4800)
    /*
     * Only paleolithic systems need this.
     */

    switch (ttybaud) {
    case 300:
	ttybaud = B300;
	break;
    case 1200:
	ttybaud = B1200;
	break;
    case 2400:
	ttybaud = B2400;
	break;
    case 4800:
	ttybaud = B4800;
	break;
    case 9600:
	ttybaud = B9600;
	break;
    case 19200:
	ttybaud = B19200;
	break;
    default:
	ttybaud = B4800;
	break;
    }
#endif

    if (cfsetispeed(&termios, ttybaud) != 0) {
	perror("cfsetispeed");
	return (-1);
    }
    if (cfsetospeed(&termios, ttybaud) != 0) {
	perror("cfsetospeed");
	return (-1);
    }
    if (tcsetattr(fd, TCSANOW, &termios) < 0) {
	perror("tcsetattr");
	return (-1);
    }
#if WANT_BLOCKING_READ
    if (fcntl(fd, F_SETFL, 0) == -1) {
	perror("fcntl: set nonblock");
    }
#endif
    return (fd);
}

void
fullwrite(int fd,
	  u_char * buf,
	  int cnt)
{
    int             ret;

    while (cnt > 0) {
	ret = write(fd, buf, cnt);
	if (ret > 0) {
	    cnt -= ret;
	    buf += ret;
	} else {
	    perror("error writing");
	    exit(1);
	}
    }
}

void
usage(void)
{
    fprintf(stderr, "dgpsip receives DGPS over IP and sends to serial port\n");
    fprintf(stderr, "Usage: dgpsip [OPTIONS]\n");
    fprintf(stderr, "  Options are:\n");
    fprintf(stderr, "    -c capture_file  saves GPS NMEA data to file\n");
    fprintf(stderr, "    -d    decode RTCM and print to stdout\n");
    fprintf(stderr, "       this turns off tty output, use -o to turn it on again\n");
    fprintf(stderr, "    -h DGPS_host  sets name of server to connect to\n");
    fprintf(stderr, "       default is dgps.wsrcc.com\n");
    fprintf(stderr, "    -p DGPS_port  sets IP port #/name to connect to\n");
    fprintf(stderr, "       default is IANA assigned port rtcm-sc104 (2101)\n");
    fprintf(stderr, "    -o serial_port  sets name of serial output device\n");
    fprintf(stderr, "       default is /dev/gps - a link to /dev/tty??\n");
    fprintf(stderr, "    -s serial_speed  sets serial output baud rate\n");
    fprintf(stderr, "       default is 4800 baud\n");
    fprintf(stderr, "    -v verblevel  turns on extra output\n");
    fprintf(stderr, "       see README for details\n");
    fprintf(stderr, "    -w turns off startup warranty message\n");
    exit(2);
}

int
tcpopen(u_char * host,
	u_char * servp)
{
    int             sockt,
                    port;
    struct hostent *hp;
    struct servent *sp;
    struct sockaddr_in rem_sockaddr_in;

    memset((char *) &rem_sockaddr_in, 0, sizeof(rem_sockaddr_in));

    if ((rem_sockaddr_in.sin_addr.s_addr = inet_addr(host)) != INADDR_NONE) {
	hp = gethostbyaddr((const char *) &rem_sockaddr_in.sin_addr,
			   sizeof(rem_sockaddr_in.sin_addr), AF_INET);
	if (hp == NULL) {
	    /*
	     * herror("gethostbyaddr"); not an error for numeric
	     * addresses
	     */
	} else {
	    bcopy(hp->h_addr, (char *) &rem_sockaddr_in.sin_addr, hp->h_length);
	}
    } else {
	hp = gethostbyname(host);
	if (hp == NULL) {
	    herror("gethostbyname");
	    return (-1);
	}
	bcopy(hp->h_addr, (char *) &rem_sockaddr_in.sin_addr, hp->h_length);
    }
    sockt = socket(AF_INET, SOCK_STREAM, 0);
    if (sockt < 0) {
	perror("socket/tcp");
	return (-1);
    }
    port = htons(atoi(servp));
    if (!port) {
	sp = getservbyname(servp, "tcp");
	if (sp == 0) {
	    if (strcmp(servp, "rtcm-sc104") == 0) {
		sp = getservent();
		if (sp) {
		    sp->s_port = htons(2101);
		}
	    }
	}
	if (sp == 0) {
	    fprintf(stderr, "unknown service %s/%s\n", servp,
		    "tcp");
	    close(sockt);
	    return (-1);
	}
	port = sp->s_port;
    }
#ifndef SOCKADDR_IN_HAS_NO_SINLEN
    rem_sockaddr_in.sin_len = sizeof(rem_sockaddr_in);
#endif
    rem_sockaddr_in.sin_family = AF_INET;
    rem_sockaddr_in.sin_port = port;

    if (connect(sockt, (const struct sockaddr *) & rem_sockaddr_in,
		sizeof(rem_sockaddr_in)) == -1) {
	if (verbose >= 5)
	    perror("connect");
	close(sockt);
	return (-1);
    }
    return (sockt);
}

/* end */
